/*************************************************
*     Command Line Decoding Routine (rdargs)     *
*************************************************/

/* Header for rdargs routine */

typedef struct {
  int presence;
  int number;
  const char *text;
} arg_result;

enum { arg_present_not, arg_present_unkeyed, arg_present_keyed };

extern int rdargs(int, char **, const char *, arg_result *);

/* End of rdargs.h */
